"""EyeOnWater API integration."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .exceptions import EyeOnWaterException
from .models import DataPoint
from .units import EOWUnits, convert_to_native, deduce_native_units

if TYPE_CHECKING:  # pragma: no cover
    from .client import Client
    from .meter_reader import MeterReader
    from .models import MeterInfo, Reading

    pass

SEARCH_ENDPOINT = "/api/2/residential/new_search"
CONSUMPTION_ENDPOINT = "/api/2/residential/consumption?eow=True"

_LOGGER = logging.getLogger(__name__)


class Meter:
    """Class represents meter state."""

    def __init__(self, reader: MeterReader, meter_info: MeterInfo) -> None:
        """Initialize the meter."""
        self.reader = reader
        self.last_historical_data: list[DataPoint] = []

        self._reading_data: Reading | None = None
        self._meter_info = meter_info
        self._reading_data = self._meter_info.reading

        self._native_unit_of_measurement = deduce_native_units(
            self._meter_info.reading.latest_read.units
        )

    @property
    def meter_uuid(self) -> str:
        """Return meter UUID."""
        return self.reader.meter_uuid

    @property
    def meter_id(self) -> str:
        """Return meter ID."""
        return self.reader.meter_id

    @property
    def native_unit_of_measurement(self) -> str:
        """Return native measurement units."""
        return self._native_unit_of_measurement.value

    async def read_meter_info(self, client: Client) -> None:
        """Read the latest meter info."""
        self._meter_info = await self.reader.read_meter_info(client)
        self._reading_data = self._meter_info.reading

    async def read_historical_data(self, client: Client, days_to_load: int) -> None:
        """Read historical data for N last days."""
        historical_data = await self.reader.read_historical_data(
            client=client, days_to_load=days_to_load
        )

        historical_data = [self._convert_to_native(dp) for dp in historical_data]

        if not self.last_historical_data:
            self.last_historical_data = historical_data
        elif (
            historical_data
            and historical_data[-1].dt > self.last_historical_data[-1].dt
        ):
            # Take newer data
            self.last_historical_data = historical_data
        elif historical_data[-1].reading == self.last_historical_data[
            -1
        ].reading and len(historical_data) > len(self.last_historical_data):
            # If it the same date - take more data
            self.last_historical_data = historical_data

    @property
    def meter_info(self) -> MeterInfo:
        """Return MeterInfo."""
        if not self._meter_info:
            msg = "Data was not fetched"
            raise EyeOnWaterException(msg)
        return self._meter_info

    @property
    def reading(self) -> DataPoint:
        """Returns the latest meter reading."""
        if not self._reading_data:
            msg = "Data was not fetched"
            raise EyeOnWaterException(msg)
        reading = self._reading_data.latest_read
        dp = DataPoint(
            dt=reading.read_time, reading=reading.full_read, unit=reading.units
        )

        return self._convert_to_native(dp)

    def _convert_to_native(self, dp: DataPoint) -> DataPoint:
        """Convert data point to meters native units"""
        native_reading = convert_to_native(
            self._native_unit_of_measurement, EOWUnits(dp.unit), dp.reading
        )
        return DataPoint(
            dt=dp.dt, reading=native_reading, unit=self._native_unit_of_measurement
        )



"""Units related tools."""
from .exceptions import EyeOnWaterUnitError
from .models import EOWUnits, NativeUnits


def deduce_native_units(read_unit: EOWUnits) -> NativeUnits:
    """Deduce native units based on oew units"""

    if read_unit in [EOWUnits.UNIT_CUBIC_METER, EOWUnits.UNIT_CM]:
        return NativeUnits.CM
    elif read_unit in [
        EOWUnits.UNIT_GAL,
        EOWUnits.UNIT_10_GAL,
        EOWUnits.UNIT_100_GAL,
        EOWUnits.UNIT_KGAL,
    ]:
        return NativeUnits.GAL
    elif read_unit in [
        EOWUnits.UNIT_CCF,
        EOWUnits.UNIT_CF,
        EOWUnits.UNIT_CUBIC_FEET,
    ]:
        return NativeUnits.CCF
    else:
        msg = f"Unsupported measurement unit: {read_unit}"
        raise EyeOnWaterUnitError(
            msg,
        )


def convert_to_native(  # noqa: C901
    native_unit: NativeUnits, read_unit: EOWUnits, value: float
) -> float:
    """Convert read units to native unit."""

    if native_unit == NativeUnits.CM:
        if read_unit in [EOWUnits.UNIT_CUBIC_METER, EOWUnits.UNIT_CM]:
            return value
        else:
            msg = f"Unsupported measurement unit: {read_unit} for native unit: {native_unit}"
            raise EyeOnWaterUnitError(msg)
    elif native_unit == NativeUnits.GAL:
        if read_unit == EOWUnits.UNIT_KGAL:
            return value * 1000
        elif read_unit == EOWUnits.UNIT_100_GAL:
            return value * 100
        elif read_unit == EOWUnits.UNIT_10_GAL:
            return value * 10
        elif read_unit == EOWUnits.UNIT_GAL:
            return value
        else:
            msg = f"Unsupported measurement unit: {read_unit} for native unit: {native_unit}"
            raise EyeOnWaterUnitError(msg)
    elif native_unit == NativeUnits.CCF:
        if read_unit in [EOWUnits.UNIT_CF, EOWUnits.UNIT_CUBIC_FEET]:
            return value / 100
        if read_unit == EOWUnits.UNIT_CCF:
            return value
        else:
            msg = f"Unsupported measurement unit: {read_unit} for native unit: {native_unit}"
            raise EyeOnWaterUnitError(msg)
    else:
        msg = f"Unsupported native unit: {native_unit}"
        raise EyeOnWaterUnitError(
            msg,
        )



"""EyeOnWater API integration."""
from __future__ import annotations

import datetime
import json
import logging
from typing import TYPE_CHECKING

from pydantic import ValidationError
import pytz

from .exceptions import EyeOnWaterAPIError, EyeOnWaterResponseIsEmpty
from .models import DataPoint, HistoricalData, MeterInfo

if TYPE_CHECKING:  # pragma: no cover
    from .client import Client

    pass

SEARCH_ENDPOINT = "/api/2/residential/new_search"
CONSUMPTION_ENDPOINT = "/api/2/residential/consumption?eow=True"

_LOGGER = logging.getLogger(__name__)


class MeterReader:
    """Class represents meter reader."""

    def __init__(self, meter_uuid: str, meter_id: str) -> None:
        """Initialize the meter."""
        self.meter_uuid = meter_uuid
        self.meter_id: str = meter_id

    async def read_meter_info(self, client: Client) -> MeterInfo:
        """Triggers an on-demand meter read and returns it when complete."""
        _LOGGER.debug("Requesting meter reading")

        query = {"query": {"terms": {"meter.meter_uuid": [self.meter_uuid]}}}
        data = await client.request(path=SEARCH_ENDPOINT, method="post", json=query)
        data = json.loads(data)
        meters = data["elastic_results"]["hits"]["hits"]
        if len(meters) > 1:
            msg = "More than one meter reading found"
            raise Exception(msg)

        try:
            meter_info = MeterInfo.parse_obj(meters[0]["_source"])
        except ValidationError as e:
            msg = f"Unexpected EOW response {e}"
            raise EyeOnWaterAPIError(msg) from e

        return meter_info

    async def read_historical_data(
        self, client: Client, days_to_load: int
    ) -> list[DataPoint]:
        """Retrieve historical data for today and past N days."""
        today = datetime.datetime.now().replace(
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
        )

        date_list = [today - datetime.timedelta(days=x) for x in range(0, days_to_load)]
        date_list.reverse()

        _LOGGER.info(
            f"requesting historical statistics for {self.meter_uuid} on {date_list}",
        )

        statistics = []

        for date in date_list:
            _LOGGER.info(
                f"requesting historical statistics for {self.meter_uuid} on {date}",
            )
            try:
                statistics += await self.read_historical_data_one_day(
                    client=client, date=date
                )
            except EyeOnWaterResponseIsEmpty:
                continue

        return statistics

    async def read_historical_data_one_day(
        self,
        client: Client,
        date: datetime.datetime,
    ) -> list[DataPoint]:
        """Retrieve the historical hourly water readings for a requested day."""
        query = {
            "params": {
                "source": "barnacle",
                "aggregate": "hourly",
                "units": "cm",  # This parameter seems to be ignored and does not affect the output values :-)
                "combine": "true",
                "perspective": "billing",
                "display_minutes": True,
                "display_hours": True,
                "display_days": True,
                "date": date.strftime("%m/%d/%Y"),
                "furthest_zoom": "hr",
                "display_weeks": True,
            },
            "query": {"query": {"terms": {"meter.meter_uuid": [self.meter_uuid]}}},
        }
        raw_data = await client.request(
            path=CONSUMPTION_ENDPOINT,
            method="post",
            json=query,
        )
        try:
            data = HistoricalData.parse_raw(raw_data)
        except ValidationError as e:
            msg = f"Unexpected EOW response {e}"
            raise EyeOnWaterAPIError(msg) from e

        key = f"{self.meter_uuid},0"
        if key not in data.timeseries:
            msg = f"Meter {key} not found"
            raise EyeOnWaterResponseIsEmpty(msg)

        timezones = data.hit.meter_timezone
        timezone = pytz.timezone(timezones[0])

        ts = data.timeseries[key].series
        statistics = []
        for d in ts:
            statistics.append(
                DataPoint(
                    dt=timezone.localize(d.date),
                    reading=d.bill_read,
                    unit=d.display_unit,
                ),
            )

        statistics.sort(key=lambda d: d.dt)

        return statistics



"""EyeOnWater API integration."""
from __future__ import annotations

from typing import TYPE_CHECKING
import urllib.parse

from .exceptions import EyeOnWaterAPIError
from .meter import Meter
from .meter_reader import MeterReader

if TYPE_CHECKING:  # pragma: no cover
    from .client import Client

DASHBOARD_ENDPOINT = "/dashboard/"
METER_UUID_FIELD = "meter_uuid"
METER_ID_FIELD = "meter_id"
INFO_PREFIX = "AQ.Views.MeterPicker.meters = "


class Account:
    """Class represents account object."""

    def __init__(
        self,
        eow_hostname: str,
        username: str,
        password: str,
    ) -> None:
        """Initialize the account."""
        self.eow_hostname = eow_hostname
        self.username = username
        self.password = password

    async def fetch_meter_readers(self, client: Client) -> list[MeterReader]:
        """List the meter readers associated with the account."""
        path = DASHBOARD_ENDPOINT + urllib.parse.quote(self.username)
        data = await client.request(path=path, method="get")

        meters = []
        lines = data.split("\n")
        for line in lines:
            if INFO_PREFIX in line:
                meter_infos = client.extract_json(line, INFO_PREFIX)
                for meter_info in meter_infos:
                    if METER_UUID_FIELD not in meter_info:
                        msg = f"Cannot find {METER_UUID_FIELD} field"
                        raise EyeOnWaterAPIError(
                            msg,
                        )

                    meter_uuid = meter_info[METER_UUID_FIELD]
                    meter_id = meter_info[METER_ID_FIELD]

                    meter = MeterReader(
                        meter_uuid=meter_uuid,
                        meter_id=meter_id,
                    )
                    meters.append(meter)

        return meters

    async def fetch_meters(self, client: Client) -> list[Meter]:
        """List the meter states associated with the account."""
        meter_readers = await self.fetch_meter_readers(client)
        meters: list[Meter] = []
        for reader in meter_readers:
            meter_info = await reader.read_meter_info(client)
            meters.append(Meter(reader, meter_info))

        return meters



"""EyeOnWater API integration."""
from __future__ import annotations

import datetime
import json
import logging
from typing import TYPE_CHECKING, Any

from tenacity import retry, retry_if_exception_type

if TYPE_CHECKING:  # pragma: no cover
    from aiohttp import ClientSession

    from .account import Account

from .exceptions import (
    EyeOnWaterAuthError,
    EyeOnWaterAuthExpired,
    EyeOnWaterException,
    EyeOnWaterRateLimitError,
)

TOKEN_EXPIRATION = datetime.timedelta(minutes=15)
AUTH_ENDPOINT = "account/signin"

_LOGGER = logging.getLogger(__name__)


class Client:
    """Class represents client object."""

    def __init__(self, websession: ClientSession, account: Account) -> None:
        """Initialize the client."""
        self.base_url = (
            "https://" + account.eow_hostname + "/" if account.eow_hostname else ""
        )
        self.username = account.username
        self.password = account.password
        self.websession = websession
        self.cookies = None
        self.authenticated = False
        self.token_expiration = datetime.datetime.now()
        self.user_agent = None

    def _update_token_expiration(self) -> None:
        self.token_expiration = datetime.datetime.now() + TOKEN_EXPIRATION

    @retry(retry=retry_if_exception_type(EyeOnWaterAuthExpired))  # type: ignore
    async def request(
        self,
        path: str,
        method: str,
        **kwargs: Any,
    ) -> str:
        """Make API calls against the eow API."""
        await self.authenticate()
        resp = await self.websession.request(
            method,
            f"{self.base_url}{path}",
            cookies=self.cookies,
            **kwargs,
        )
        if resp.status == 403:
            _LOGGER.error("Reached ratelimit")
            msg = "Reached ratelimit"
            raise EyeOnWaterRateLimitError(msg)
        elif resp.status == 401:
            _LOGGER.debug("Authentication token expired; requesting new token")
            self.authenticated = False
            await self.authenticate()
            raise EyeOnWaterAuthExpired

        # Since API call did not return a 400 code, update the token_expiration.
        self._update_token_expiration()

        data: str = await resp.text()

        if resp.status != 200:
            _LOGGER.error(f"Request failed: {resp.status} {data}")
            msg = f"Request failed: {resp.status} {data}"
            raise EyeOnWaterException(msg)

        return data

    async def authenticate(self) -> None:
        """Authenticate the client."""
        if not self.is_token_valid:
            _LOGGER.debug("Requesting login token")

            resp = await self.websession.request(
                "POST",
                f"{self.base_url}{AUTH_ENDPOINT}",
                data={
                    "username": self.username,
                    "password": self.password,
                },
            )

            # if "dashboard" not in str(resp.url):

            if resp.status == 400:
                msg = f"Username or password was not accepted by {self.base_url}"
                raise EyeOnWaterAuthError(msg)

            if resp.status == 403:
                msg = "Reached ratelimit"
                raise EyeOnWaterRateLimitError(msg)

            self.cookies = resp.cookies
            self._update_token_expiration()
            self.authenticated = True
            _LOGGER.debug("Successfully retrieved login token")

    def extract_json(self, line: str, prefix: str) -> list[dict[str, Any]]:
        """Extract JSON response."""
        line = line[line.find(prefix) + len(prefix) :]
        line = line[: line.find(";")]
        return json.loads(line)  # type: ignore

    @property
    def is_token_valid(self) -> bool:
        """Validate the token."""
        if self.authenticated or (datetime.datetime.now() < self.token_expiration):
            return True

        return False



"""EyeOnWater API integration."""
from __future__ import annotations

from .account import Account  # noqa
from .client import Client  # noqa
from .exceptions import *  # noqa
from .exceptions import EyeOnWaterAuthError  # noqa
from .meter import Meter  # noqa
from .meter_reader import MeterReader  # noqa
from .models import *  # noqa
from .units import *  # noqa



# ruff: noqa

from __future__ import annotations

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, Field

from .units import EOWUnits


class Service(BaseModel):
    # Optional fields
    class_code_normalized: Optional[str] = None
    route: Optional[str] = None
    active: Optional[bool] = None
    class_code: Optional[str] = None
    service_type: Optional[str] = None
    service_id: Optional[str] = None
    start_date: Optional[datetime] = None
    service_point_uuid: Optional[str] = None
    cycle: Optional[str] = None


class Location(BaseModel):
    # Optional fields
    city: Optional[str] = None
    parcel_number: Optional[str] = None
    location_name: Optional[str] = None
    parity: Optional[str] = None
    country: Optional[str] = None
    route: Optional[str] = None
    geocode_status: Optional[int] = None
    zip_code: Optional[str] = None
    longitude: Optional[str] = None
    display_address_3: Optional[str] = None
    state: Optional[str] = None
    location_uuid: Optional[str] = None
    county_name: Optional[str] = None
    latitude: Optional[str] = None
    display_address_2: Optional[str] = None
    location_id: Optional[str] = None
    display_address: Optional[str] = None
    display_street_name: Optional[str] = None
    cycle: Optional[str] = None


class LeakAlert(BaseModel):
    # Optional fields
    alert_type: Optional[str] = None
    name: Optional[str] = None
    residential_user_name: Optional[str] = None
    date_updated: Any
    alert_uuid: Optional[str] = None
    state: Optional[str] = None
    date_created: Optional[str] = None
    creator_user_uuid: Optional[str] = None


class Alerts(BaseModel):
    # Optional fields
    leak_alert: Optional[LeakAlert] = None


class AccountInfo(BaseModel):
    # Optional fields
    status: Optional[str] = None
    first_name: Optional[str] = None
    billing_address: Optional[str] = None
    billing_city: Optional[str] = None
    account_uuid: Optional[str] = None
    billing_address_2: Optional[str] = None
    full_name: Optional[str] = None
    email: Optional[str] = None
    phone: Optional[str] = None
    portal_status: Optional[str] = None
    last_name: Optional[str] = None
    account_id: Optional[str] = None
    class_code: Optional[str] = None
    billing_address_3: Optional[str] = None
    person_id: Optional[str] = None
    date_created: Optional[str] = None
    account_billing_cycle: Optional[str] = None
    billing_zip_code: Optional[str] = None
    billing_country: Optional[str] = None
    billing_state: Optional[str] = None
    eyeonwater: Optional[str] = None


class SensorsAvailable(BaseModel):
    # Optional fields
    types: Optional[list[str]] = None


class Battery(BaseModel):
    # Optional fields
    register_: Optional[int] = Field(None, alias="register")
    level: Optional[int] = None
    quality: Optional[str] = None
    thresh12mo: Optional[int] = None
    time: Optional[datetime] = None


class Pwr(BaseModel):
    # Optional fields
    level: Optional[int] = None
    register_: Optional[int] = Field(None, alias="register")
    signal_7days: Optional[int] = None
    signal_strength: Optional[int] = None
    time: Optional[datetime] = None
    quality: Optional[str] = None
    signal_30days: Optional[int] = None


class Notes(BaseModel):
    # Optional fields
    count: Optional[int] = None


class Flow(BaseModel):
    # Optional fields
    this_week: Optional[float] = None
    months_updated: Optional[str] = None
    last_month: Optional[float] = None
    last_year_last_month_ratio: Optional[float] = None
    last_year_last_month: Optional[float] = None
    delta_positive: Optional[float] = None
    time: Optional[datetime] = None
    time_positive: Optional[str] = None
    last_month_ratio: Optional[float] = None
    last_week_avg: Optional[float] = None
    last_year_this_month_ratio: Optional[float] = None
    delta: Optional[float] = None
    this_month: Optional[float] = None
    week_ratio: Optional[float] = None
    weeks_updated: Optional[str] = None
    last_year_this_month: Optional[float] = None
    last_week: Optional[float] = None
    this_week_avg: Optional[float] = None


class ActiveFlags(BaseModel):
    # Optional fields
    active_flags: Optional[list[str]] = None
    time: Optional[datetime] = None


class MeterData(BaseModel):
    # Optional fields
    sensors_available: Optional[SensorsAvailable] = None
    has_endpoint: Optional[bool] = None
    install_date: Optional[datetime] = None
    meter_uuid: Optional[str] = None
    fluid_type: Optional[str] = None
    timezone: Optional[str] = None
    firmware_version: Optional[str] = None
    communication_security: Optional[str] = None
    meter_size_unit: Optional[str] = None
    cell_type: Optional[str] = None
    geocode_status: Optional[int] = None
    geo: Optional[str] = None
    last_read_time: Optional[datetime] = None
    note: Optional[str] = None
    battery: Optional[Battery] = None
    endpoint_status: Optional[str] = None
    alert_code: Optional[int] = None
    gas_pressure_compensation: Optional[float] = None
    service_type: Optional[str] = None
    serial_number: Optional[str] = None
    type_: Optional[str] = Field(None, alias="type")
    pwr: Optional[Pwr] = None
    endpoint_connector: Optional[str] = None
    communication_seconds: Optional[int] = None
    meter_size_desc: Optional[str] = None
    is_compound: Optional[str] = None
    latitude: Optional[str] = None
    typical_read_method: Optional[str] = None
    endpoint_type: Optional[str] = None
    meter_id: Optional[str] = None
    last_communication_time: Optional[datetime] = None
    pit_type: Optional[str] = None
    meter_spec_uuid: Optional[str] = None
    manufacturer: Optional[str] = None
    has_valve: Optional[bool] = None
    has_sensor: Optional[bool] = None
    gas_sub_count: Optional[int] = None
    notes: Optional[Notes] = None
    meter_size: Optional[float] = None
    flow: Optional[Flow] = None
    longitude: Optional[str] = None
    flags: Optional[ActiveFlags] = None
    model: Optional[str] = None
    sequence_number: Optional[int] = None


class ServiceAgreement(BaseModel):
    # Optional fields
    service_agreement_uuid: Optional[str] = None
    start_date: Optional[datetime] = None


class LatestRead(BaseModel):
    # Mandatory fields
    full_read: float
    units: EOWUnits

    # Optional fields
    bill_read: Optional[str] = None
    bill_display_units: Optional[EOWUnits] = None
    read_time: datetime
    has_endpoints: Optional[bool] = None
    method: Optional[str] = None


class Timeslots(BaseModel):
    # Optional fields
    weekend: Optional[list[int]] = None
    weekday: Optional[list[int]] = None


class Encoder(BaseModel):
    # Optional fields
    time: Optional[datetime] = None
    dials: Optional[int] = None
    register_id: Optional[str] = None
    totalizer: Optional[int] = None


class Flags(BaseModel):
    # Mandatory fields (used by EOW HA integration)
    empty_pipe: bool = Field(..., alias="EmptyPipe")
    leak: bool = Field(..., alias="Leak")
    cover_removed: bool = Field(..., alias="CoverRemoved")
    tamper: bool = Field(..., alias="Tamper")
    reverse_flow: bool = Field(..., alias="ReverseFlow")
    low_battery: bool = Field(..., alias="LowBattery")
    battery_charging: bool = Field(..., alias="BatteryCharging")

    # Optional fields
    forced: Optional[bool] = Field(None, alias="Forced")
    magnetic_tamper: Optional[bool] = Field(None, alias="MagneticTamper")
    encoder_no_usage: Optional[bool] = Field(None, alias="EncoderNoUsage")
    encoder_temperature: Optional[bool] = Field(None, alias="EncoderTemperature")
    encoder_reverse_flow: Optional[bool] = Field(None, alias="EncoderReverseFlow")
    reading_changed: Optional[bool] = Field(None, alias="ReadingChanged")
    programming_changed: Optional[bool] = Field(None, alias="ProgrammingChanged")
    encoder_exceeding_max_flow: Optional[bool] = Field(
        None, alias="EncoderExceedingMaxFlow"
    )
    water_temperature_sensor_error: Optional[bool] = Field(
        None, alias="WaterTemperatureSensorError"
    )
    oscillator_failure: Optional[bool] = Field(None, alias="OscillatorFailure")
    encoder_sensor_error: Optional[bool] = Field(None, alias="EncoderSensorError")
    encoder_leak: Optional[bool] = Field(None, alias="EncoderLeak")
    water_pressure_sensor_error: Optional[bool] = Field(
        None, alias="WaterPressureSensorError"
    )
    min_max_invalid: Optional[bool] = Field(None, alias="MinMaxInvalid")
    end_of_life: Optional[bool] = Field(None, alias="EndOfLife")
    encoder_dial_change: Optional[bool] = Field(None, alias="EncoderDialChange")
    no_usage: Optional[bool] = Field(None, alias="NoUsage")
    device_alert: Optional[bool] = Field(None, alias="DeviceAlert")
    endpoint_reading_missed: Optional[bool] = Field(None, alias="EndpointReadingMissed")
    encoder_removal: Optional[bool] = Field(None, alias="EncoderRemoval")
    profile_read_error: Optional[bool] = Field(None, alias="ProfileReadError")
    encoder_programmed: Optional[bool] = Field(None, alias="EncoderProgrammed")
    time: Optional[datetime] = None
    encoder_magnetic_tamper: Optional[bool] = Field(None, alias="EncoderMagneticTamper")
    meter_temperature_sensor_error: Optional[bool] = Field(
        None, alias="MeterTemperatureSensorError"
    )


class Reading(BaseModel):
    # Mandatory fields
    flags: Flags
    latest_read: LatestRead

    # Optional fields
    battery: Optional[Battery] = None
    customer_uuid: Optional[str] = None
    aggregation_seconds: Optional[int] = None
    last_communication_time: Optional[datetime] = None
    firmware_version: Optional[str] = None
    communication_security: Optional[str] = None
    meter_size_desc: Optional[str] = None
    barnacle_uuid: Optional[str] = None
    unit: Optional[EOWUnits] = None
    customer_name: Optional[str] = None
    cell_type: Optional[str] = None
    pi_status: Optional[str] = None
    second_carrier: Optional[bool] = None
    input_config: Optional[str] = None
    meter_size_unit: Optional[str] = None
    endpoint_status: Optional[str] = None
    gas_pressure_compensation: Optional[float] = None
    serial_number: Optional[str] = None
    activated_on: Optional[str] = None
    sim_vendor: Optional[str] = None
    pwr: Optional[Pwr] = None
    communication_seconds: Optional[int] = None
    cell_endpoint_name: Optional[str] = None
    rf_communication: Optional[bool] = None
    low_read_limit: Optional[int]
    utility_use_1: Optional[str] = None
    timeslots: Optional[Timeslots] = None
    encoder: Optional[Encoder] = None
    endpoint_type: Optional[str] = None
    utility_use_2: Optional[str] = None
    multiplier: Optional[str] = None
    sim_type: Optional[str] = None
    register_number: Optional[str] = None
    wired_interface: Optional[str] = None
    endpoint_install_date: Optional[datetime] = None
    high_read_limit: Optional[int] = None
    gas_sub_count: Optional[int] = None
    billing_number: Optional[str] = None
    meter_size: Optional[float] = None
    flow: Optional[Flow] = None
    hardware_version: Optional[str] = None
    connector_type: Optional[str] = None
    model: Optional[str] = None
    resolution: Optional[float] = None


class Conditions(BaseModel):
    # Optional fields
    increasing: Optional[bool] = None
    decreasing: Optional[bool] = None


class EndpointTemperature(BaseModel):
    # Optional fields
    latest_average: Optional[float] = None
    last_reported: Optional[str] = None
    seven_day_min: Optional[str] = None
    seven_day_average: Optional[str] = None
    seven_day_max: Optional[str] = None
    sensor_uuid: Optional[str] = None
    conditions: Optional[Conditions] = None


class Sensors(BaseModel):
    # Optional fields
    endpoint_temperature: Optional[EndpointTemperature] = None


class Utility(BaseModel):
    # Optional fields
    fluid_barrel_billing_unit: Optional[str] = None
    cm_billing_unit: Optional[str] = None
    cf_billing_unit: Optional[str] = None
    type_: Optional[str] = Field(None, alias="type")
    utility_name: Optional[str] = None
    eow_service_selector: Optional[str] = None
    gas_cf_billing_unit: Optional[str] = None
    eow_type: Optional[str] = None
    oil_barrel_billing_unit: Optional[str] = None
    date_created: Optional[str] = None
    gal_billing_unit: Optional[str] = None
    gas_cm_billing_unit: Optional[str] = None
    utility_uuid: Optional[str] = None
    imp_billing_unit: Optional[str] = None


class User(BaseModel):
    # Optional fields
    user_uuid: Optional[str] = None
    user_name: Optional[str] = None
    date_created: Optional[str] = None


class Groups(BaseModel):
    # Optional fields
    irrigation: Optional[str] = None
    continuous_flow: Optional[str] = None
    is_irrigatable: Optional[str] = None
    disable_valve_shutoff: Optional[str] = None


class MeterInfo(BaseModel):
    # Mandatory fields
    reading: Reading = Field(..., alias="register_0")

    # Optional fields
    sensors: Optional[Sensors] = None
    utility: Optional[Utility] = None
    updated: Optional[int] = None
    last_updated: Optional[str] = None
    service: Optional[Service] = None
    location: Optional[Location] = None
    alerts: Optional[Alerts] = None
    account: Optional[AccountInfo] = None
    meter: Optional[MeterData] = None
    service_agreement: Optional[ServiceAgreement] = None
    version: Optional[str] = None
    user: Optional[User] = None
    groups: Optional[Groups] = None



"""EOW Client data models."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import datetime


@dataclass
class DataPoint:
    """One data point representation."""

    dt: datetime
    reading: float
    unit: str



"""Supported EOW units."""

from __future__ import annotations

from enum import Enum


class EOWUnits(str, Enum):
    """Enum of supported EOW units."""

    UNIT_GAL = "GAL"
    UNIT_100_GAL = "100 GAL"
    UNIT_10_GAL = "10 GAL"
    UNIT_CF = "CF"
    UNIT_CUBIC_FEET = "CUBIC_FEET"
    UNIT_CCF = "CCF"
    UNIT_KGAL = "KGAL"
    UNIT_CM = "CM"
    UNIT_CUBIC_METER = "CUBIC_METER"


class NativeUnits(str, Enum):
    """Enum of supported native units."""

    GAL = "gal"
    CCF = "cf"
    CM = "cm"



"""Exceptions of EOW Client."""


class EyeOnWaterException(Exception):
    """Base exception for more specific exceptions to inherit from."""


class EyeOnWaterAuthError(EyeOnWaterException):
    """Exception for authentication failures.

    Either wrong username or wrong password.
    """


class EyeOnWaterRateLimitError(EyeOnWaterException):
    """Exception for reaching the ratelimit.

    Either too many login attempts or too many requests.
    """


class EyeOnWaterAuthExpired(EyeOnWaterException):
    """Exception for when a token is no longer valid."""


class EyeOnWaterAPIError(EyeOnWaterException):
    """General exception for unknown API responses."""


class EyeOnWaterResponseIsEmpty(EyeOnWaterException):
    """API answered correct but there is not content to parse."""


class EyeOnWaterUnitError(EyeOnWaterException):
    """Exception for units related errors."""



# ruff: noqa

from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field

from .units import EOWUnits


class Register0EncoderItem(BaseModel):
    dials: Optional[int] = None


class Hit(BaseModel):
    # Mandatory fields
    meter_timezone: list[str] = Field(..., alias="meter.timezone")

    # Optional fields
    meter_communication_seconds: Optional[list[int]] = Field(
        None, alias="meter.communication_seconds"
    )
    register_0_encoder: Optional[list[Register0EncoderItem]] = Field(
        None, alias="register_0.encoder"
    )
    location_location_uuid: Optional[list[str]] = Field(
        None, alias="location.location_uuid"
    )
    meter_fluid_type: Optional[list[str]] = Field(None, alias="meter.fluid_type")
    meter_meter_id: Optional[list[str]] = Field(None, alias="meter.meter_id")
    account_full_name: Optional[list[str]] = Field(None, alias="account.full_name")
    meter_meter_uuid: Optional[list[str]] = Field(None, alias="meter.meter_uuid")
    meter_has_endpoint: Optional[list[bool]] = Field(None, alias="meter.has_endpoint")
    meter_serial_number: Optional[list[str]] = Field(None, alias="meter.serial_number")
    account_account_id: Optional[list[str]] = Field(None, alias="account.account_id")
    location_location_name: Optional[list[str]] = Field(
        None, alias="location.location_name"
    )
    service_service_id: Optional[list[str]] = Field(None, alias="service.service_id")
    register_0_serial_number: Optional[list[str]] = Field(
        None, alias="register_0.serial_number"
    )
    utility_utility_uuid: Optional[list[str]] = Field(
        None, alias="utility.utility_uuid"
    )
    account_account_uuid: Optional[list[str]] = Field(
        None, alias="account.account_uuid"
    )


class Params(BaseModel):
    # Optional fields
    start_date_utc: Optional[float] = None
    date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    end_date_utc: Optional[float] = None
    compare: Optional[bool] = None
    read_type: Optional[str] = None
    start_date: Optional[datetime] = None
    end_date_tz: Optional[str] = None
    aggregate: Optional[str] = None
    aggregate_group: Optional[str] = None
    perspective: Optional[str] = None
    units: Optional[EOWUnits] = None
    start_date_tz: Optional[str] = None
    adjust_to: Optional[bool] = None
    combine_group: Optional[bool] = None


class Series(BaseModel):
    # Mandatory fields
    date: datetime
    display_unit: EOWUnits
    bill_read: float

    # Optional fields
    end_date: Optional[datetime] = None
    meter_uuid: Optional[int] = None
    value: Optional[float] = None
    start_date: Optional[datetime] = None
    register_number: Optional[int] = None
    estimated: Optional[int] = None
    raw_read: Optional[int] = None
    unit: Optional[EOWUnits] = None


class Legend(BaseModel):
    # Optional fields
    supply_zone_id: Optional[str] = None
    location_name: Optional[str] = None
    account_id: Optional[str] = None
    demand_zone_id: Optional[str] = None
    meter_id: Optional[str] = None
    full_name: Optional[str] = None
    serial_number: Optional[str] = None


class TimeSerie(BaseModel):
    # Mandatory fields
    series: list[Series]

    # Optional fields
    legend: Optional[Legend] = None


class HistoricalData(BaseModel):
    # Mandatory fields
    hit: Hit
    timeseries: dict[str, TimeSerie]

    # Optional fields
    min_chart_aggregation: Optional[str] = None
    params: Optional[Params] = None
    timezone: Optional[str] = None
    min_aggregation_seconds: Optional[int] = None
    annotations: Optional[list[str]] = None
